#
# Copyright (c) 1998 Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#       This product includes software developed by the Computer Systems
#       Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# ------------
#
# Filename: ui-activesourcemgr.tcl
#   -- Created on Mon Apr 20 1998
#   -- Author: Cynthia Romer <cromer@cs.berkeley.edu>
#
#  @(#) $Header: /usr/src/mash/repository/mash/mash-1/tcl/vic/ui-activesourcemgr.tcl,v 1.4 1998/10/16 20:56:43 yatin Exp $
#

import Observer ActiveSource UserWindow Configuration

#
# ActiveSourceManager is used to dispatch and manage ActiveSource
# objects.  It is an Observer of VideoAgent.
#
Class ActiveSourceManager -superclass {Observer} -configuration {
	tile 1 
}

#
# Tile out in a grid the video of the ActiveSources followed by the
# provided <i>videoAgent</i>.  If <i>list_direction</i> is "horizontal",
# the <i>tile</i> resource option indicates the initial number of rows
# for the grid.  If <i>list_direction</i> is "vertical", the <i>tile</i>
# resource option indicates the initial number of columns for the grid.
#
ActiveSourceManager public init {ui w videoAgent list_direction localChannel} {
	$self next 

	$self instvar ui_ videoAgent_ localChannel_
	set ui_ $ui
	set videoAgent_ $videoAgent
	set localChannel_ $localChannel

	$self instvar curcol_ currow_ ncol_ nrow_ list_direction_
	set curcol_ 0
	set currow_ 0
	set list_direction_ $list_direction
	set ncol_ [$self get_option tile]
	set nrow_ [$self get_option tile]

	# let this ActiveSourceManager be an observer of the VideoAgent so it can catch "trigger_format" and "activate" and "deactivate"
	$videoAgent_ attach $self	
}

#
ActiveSourceManager public init_grid {w} {
	$self instvar grid_ label_

	frame $w
	set grid_ $w.grid
	frame $grid_
	set label_ $w.label
	label $label_ -text "Waiting for video..."

	pack $label_ -anchor c -expand 1 -side left -fill both
}

#
# The provided ActiveSource, <i>as</i>, is added to the active_() instvar (an array of active senders indexed by <i>src</i>).
# If the array was previously empty, the $label_ is removed and the $grid_ is packed in its place.
#
ActiveSourceManager public add_active { as src } {
	$self instvar active_ grid_ label_
	set active_($src) $as
	if { [array size active_] == 1 } {
		pack forget $label_
		pack $grid_ -expand 1 -fill x -anchor n 
	}
}

#
# The specified <i>src</i> is removed from the active_() instvar (an array of active senders indexed by src).
# If this was the last element of the array, the $grid_ is removed and the $label_ is packed in its place.
#
ActiveSourceManager public rm_active src {
	$self instvar active_ grid_ label_
	unset active_($src)
	if { [array size active_] == 0 } {
		pack forget $grid_
		pack $label_ -anchor c -expand 1 -side left
	}
}

#
# Return a list of the active sources.
#
ActiveSourceManager public active-sources {} {
	$self instvar active_
	return [array names active_]
}

#
# Update the row and column variables to indicate an addition to the
# grid of thumbnails of ActiveSources.  If the <i>list_direction_</i> is
# "vertical", additions to the grid are placed at the end of the first
# non-full column.  When a column becomes full, a new column is begun,
# filled from top to bottom.  If the <i>list_direction_</i> is
# "horizontal", additions to the grid are placed at the end of the first
# non-full row.  When a row becomes full, a new row is begun, filled
# from left to right.
#
ActiveSourceManager public bump { } {
	$self instvar curcol_ currow_ list_direction_

	if { $list_direction_ == "vertical" } {	
		$self instvar ncol_
		incr curcol_
		if { $curcol_ == $ncol_ } {
			set curcol_ 0
			incr currow_
		}
	} else {
		$self instvar nrow_
		incr currow_
		if { $currow_ == $nrow_ } {
			set currow_ 0
			incr curcol_
		}
	}
}

#
# If the <i>list_direction_</i> is "vertical", reformat the grid of
# video thumbnails into <i>n</i> columns.  If the <i>list_direction_</i>
# is "horizontal", reformat the grid of video thumbnails into <i>n</i>
# rows.
#
ActiveSourceManager public redecorate n {
	$self instvar curcol_ currow_ list_direction_
	set curcol_ 0
	set currow_ 0

	if { $list_direction_ == "vertical" } {
		$self instvar ncol_
		set ncol_ $n
	} else {
		$self instvar nrow_
		set nrow_ $n
	}
		
	$self instvar grid_
	if ![info exists grid_] {
		return
	}
	
	foreach src [$self active-sources] {
		grid $grid_.$src -row $currow_ -column $curcol_ -sticky we
		if { $list_direction_ == "vertical" } {	
			grid columnconfigure $grid_ $curcol_ -weight 1
		} else {
			grid rowconfigure $grid_ $currow_ -weight 1
		}
		$self bump
	}
}


#
# Add a <i>src</i> to the active senders list.  E.g., make a postage
# stamp window appear, stats, etc. so that the user can select
# the video stream.
#
ActiveSourceManager public activate src {
	#
	# give the VideoAgent a chance to create and install
	# a decoder and for that decoder to see a packet so it can
	# determine the output geometry and color decimation.
	# we shouldn't have to do this (e.g., resize will
	# take care of a geometry change), but currently
	# decoders can't trigger a renderer realloation
	# when the decimation changes.XXX fix this
	# Note that trigger_format allocates the decoder object
	# is called before the event loop goes idle
	#
	after idle "$self really_activate $src"
}

#
# Create an ActiveSource, $as, within the widget, $grid_.$src, out of the provided <i>src</i>.
# (This entails dropping a postage-stamp-sized VideoWidget decoding the $src into the $grid_.$src widget.)
# The widget, $grid.$src, is subsequently positioned within $grid_ by the grid manager in
# the row specified by instvar $currow_ and the column specified by instvar $curcol_.
# The column and row variables are updated by the bump method to reflect this addition to the grid.
# Add this ActiveSource, $as, to the instvar array active_() (the array of active senders).
# Update the decoder for the specified <i>src</i>.
#
ActiveSourceManager public really_activate src {
	$self instvar grid_ curcol_ currow_ ui_ list_direction_ localChannel_

	set as [new ActiveSource $self $ui_ $grid_.$src $src $localChannel_]

	grid $grid_.$src -row $currow_ -column $curcol_ -sticky we

	if { $list_direction_ == "vertical" } {	
		grid columnconfigure $grid_ $curcol_ -weight 1
	} else {
		grid rowconfigure $grid_ $currow_ -weight 1
	}

	$ui_ update_decoder $src
	$self bump

	#
	# Someone became activated, so we have to change
	# the switchable menu to include this participant
	#
	Switcher rebuild_switch_list_menu

	# XXX this is a temporary hack while cindy works on the VideoBox class... 
	if {[$ui_ info class] == "MuiUI"} {
		$ui_ maybe_switch_in $src
	}
}

#
# Remove a <i>src</i> from the active senders list. 
#
ActiveSourceManager public deactivate {src} {
	$self instvar active_
	if [info exists active_($src)] {
		set as $active_($src)
		set L [$as user-windows]
		foreach uw $L {
			#XXX should check if we're voice-switched
			# and if so, bump window
			delete $uw
		}
		# detach any we missed by deleting user-windows
		$as detach-windows
		$as delete-decoder-window
	}

	$self instvar grid_
	set w $grid_.$src
	if [winfo exists $w] {
		grid forget $w
		destroy $w
		$self rm_active $src
	}

	$src handler ""

	#
	# Someone became de-activated, so we have to change
	# the switchable menu to un-include this participant
	#
	Switcher rebuild_switch_list_menu

	#XXX
	global ftext btext ltext fpshat bpshat lhat shat
	unset ftext($src)
	unset btext($src)
	unset ltext($src)
	unset fpshat($src)
	unset bpshat($src)
	unset lhat($src)
	unset shat($src)
}

#
# Dispatch focus method on switcher windows.
#
ActiveSourceManager public focus_speaker { infoname msg } {
	foreach s [$self active-sources] {
		if { [$s sdes cname] == $msg } {
		        Switcher focus $s
		} 
        }
}

#
# Update the name of the <i>src</i> for the UserWindows in which it appears.
#
ActiveSourceManager public change_name {src} {
	set name [$src sdes name]
	# update viewing window names to reflect new name
	$self instvar active_
	if [info exists active_($src)] {
		set as $active_($src)
		foreach uw [$as user-windows] {
			if { [$uw attached-source] == "$as" } {
				$uw set-name $name
			}
		}
	}
	
	# Someone's name has been altered , so we have to change
	# the switchable menu to include this participant
	#
	Switcher rebuild_switch_list_menu
}

#
ActiveSourceManager public trigger_format {src} {
	$self instvar active_ videoAgent_ ui_
	
	if ![info exists active_($src)] {
		#
		# if we get a change format before really_activate
		# was called (i.e., so we don't even have a thumbnail yet),
		# don't do anything
		#
		return
	}

	set as $active_($src)

	set L [$as user-windows]
	$as detach-windows
	#XXX
	set extoutList [extout_detach_src $src]

	set d [$videoAgent_ reactivate $src]

	$ui_ update_decoder $src
	global colorbutton
	$d color $colorbutton($src)

	foreach uw $L {
		$as attach-window $uw
		[$uw video-widget] redraw
	}
	$as attach-thumbnail
	#XXX
	extout_attach_src $src $extoutList
}

#
# Called when the video stream state changes in a way that would
# affect the choice of renderer.  For example, when a jpeg stream
# changes from type-0 to type-1 we might have to revert from
# hardware to software decoding, or we might have to reallocate
# a 422 renderer as a 411 renderer.  This never needs to happen
# for most stream types (i.e., because the decimation factor is fixed).
#
ActiveSourceManager public decoder_changed {src} {
	$self instvar active_
	if ![info exists active_($src)] {
		return
	}
	#XXX redundant with trigger_format
	set as $active_($src)

	set L [$as user-windows]
	$as detach-windows
	#XXX
	set extoutList [extout_detach_src $src]
	foreach w $iw {
		$as attach-window $uw
		$w redraw
	}
	#XXX
	extout_attach_src $src $extoutList
	return
}


ActiveSourceManager public trigger_format_all { } {
	foreach s [$self active-sources] {
		$self trigger_format $s
	}
}


ActiveSourceManager public get_activesource src {
	$self instvar active_
	if [info exists active_($src)] {
		return $active_($src)
	} else {
		return ""
	}
}
